from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass, field
from pathlib import Path
from typing import Generator, TypeAlias

import geopandas as gpd
import pandas as pd
from shapely import Polygon

from .blocks import read_blocks
from .filter import filter_groups
from .nodes import consolidate_nodes
from .proto import HeaderBlock, PrimitiveBlock
from .relations import consolidate_relations
from .unpacked import BaseGroup, NodesGroup, RelationGroup, WayGroup
from .ways import consolidate_ways

__all__ = ["OSMFile"]

BBox: TypeAlias = tuple[float, float, float, float] | Polygon
ReferenceDict: TypeAlias = defaultdict[str, set[int]]


def _unpack_primitive_block(
    block: PrimitiveBlock,
) -> Generator[BaseGroup, None, None]:
    string_table: list[str] = [x.decode("utf-8") for x in block.stringtable.s]

    for group in block.primitivegroup:
        # each group contains only one field at a time, where the fields can be
        # nodes, dense, ways, relations or changesets

        if len(group.nodes) > 0:
            raise NotImplementedError()

        if len(group.dense.id) > 0:
            yield NodesGroup.from_dense_group(
                group,
                string_table,
                granularity=block.granularity,
                lat_offset=block.lat_offset,
                lon_offset=block.lon_offset,
            )

        if len(group.ways) > 0:
            yield WayGroup.from_primitive_group(group, string_table)

        if len(group.relations) > 0:
            yield RelationGroup.from_primitive_group(group, string_table)


def _read_and_unpack_groups(
    file_iterator: Generator[bytes, None, None],
) -> Generator[BaseGroup, None, None]:
    """Parse all nodes from a file-iterator."""
    for block in file_iterator:
        yield from _unpack_primitive_block(PrimitiveBlock.FromString(block))


@dataclass
class OSMFile:
    nodes: list[NodesGroup] = field(default_factory=list)
    ways: list[WayGroup] = field(default_factory=list)
    relations: list[RelationGroup] = field(default_factory=list)

    # protected property that is used to store the arguments to filter
    # for later use during consolidation, since pre-consolidation filtering
    # can leave non-matching geometries that are referenced by some way or relation
    _filter: set[str] | None = None

    @classmethod
    def from_file(cls, fp: Path | str) -> OSMFile:
        if isinstance(fp, str):
            fp = Path(fp)

        nodes: list[NodesGroup] = []
        ways: list[WayGroup] = []
        relations: list[RelationGroup] = []

        with open(fp, "rb") as f:
            file_iterator = read_blocks(f)

            # fixme: do something with header block here
            _: HeaderBlock = HeaderBlock.FromString(next(file_iterator))

            for group in _read_and_unpack_groups(file_iterator):
                match group:
                    case NodesGroup():
                        nodes.append(group)
                    case WayGroup():
                        ways.append(group)
                    case RelationGroup():
                        relations.append(group)

        return cls(nodes, ways, relations)

    def filter(self, *, tags: set[str]) -> OSMFile:
        self.relations, references = filter_groups(self.relations, tags=tags)
        self.ways, references = filter_groups(
            self.ways, tags=tags, references=references
        )
        self.nodes, _ = filter_groups(self.nodes, tags=tags, references=references)

        self._filter = tags
        return self

    def _consolidate_nodes(self) -> gpd.GeoDataFrame:
        _node_parts = [
            consolidate_nodes(nodes) for nodes in self.nodes if not nodes.is_empty()
        ]
        if len(_node_parts) > 0:
            return pd.concat(_node_parts)
        else:
            raise ValueError("Nothing to consolidate.")

    def _consolidate_ways(self, *, nodes: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
        _way_parts = [
            consolidate_ways(ways, nodes=nodes)
            for ways in self.ways
            if not ways.is_empty()
        ]

        if len(_way_parts) > 0:
            return pd.concat(_way_parts)
        else:
            return gpd.GeoDataFrame()

    def _consolidate_relations(
        self, *, nodes: gpd.GeoDataFrame, ways: gpd.GeoDataFrame
    ) -> gpd.GeoDataFrame:
        _relation_parts = [
            consolidate_relations(relations, ways=ways, nodes=nodes)
            for relations in self.relations
            if not relations.is_empty()
        ]

        if len(_relation_parts) > 0:
            return pd.concat(_relation_parts)
        else:
            return gpd.GeoDataFrame()

    def consolidate(self) -> gpd.GeoDataFrame:
        nodes = self._consolidate_nodes()
        ways = self._consolidate_ways(nodes=nodes)
        relations = self._consolidate_relations(nodes=nodes, ways=ways)

        gdf = pd.concat([nodes, ways, relations])

        if self._filter is not None:
            # filter for rows that match a filter category
            gdf = gdf[gdf[list(self._filter)].notna().any(axis=1)]

            # drop columns that became all NA from filtering
            gdf = gdf[gdf.columns[~gdf.isna().all()]]

        return gdf


#  header_bbox = box(
#      header_block.bbox.bottom * 1e-9,
#      header_block.bbox.left * 1e-9,
#      header_block.bbox.top * 1e-9,
#      header_block.bbox.right * 1e-9,
#  )
#
#  if bounding_box is not None and not intersects(header_bbox, bounding_box):



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: osmformat.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x0fosmformat.proto\x12\x06OSMPBF"\x87\x02\n\x0bHeaderBlock\x12 \n\x04\x62\x62ox\x18\x01 \x01(\x0b\x32\x12.OSMPBF.HeaderBBox\x12\x19\n\x11required_features\x18\x04 \x03(\t\x12\x19\n\x11optional_features\x18\x05 \x03(\t\x12\x16\n\x0ewritingprogram\x18\x10 \x01(\t\x12\x0e\n\x06source\x18\x11 \x01(\t\x12%\n\x1dosmosis_replication_timestamp\x18  \x01(\x03\x12+\n#osmosis_replication_sequence_number\x18! \x01(\x03\x12$\n\x1cosmosis_replication_base_url\x18" \x01(\t"F\n\nHeaderBBox\x12\x0c\n\x04left\x18\x01 \x02(\x12\x12\r\n\x05right\x18\x02 \x02(\x12\x12\x0b\n\x03top\x18\x03 \x02(\x12\x12\x0e\n\x06\x62ottom\x18\x04 \x02(\x12"\xd2\x01\n\x0ePrimitiveBlock\x12(\n\x0bstringtable\x18\x01 \x02(\x0b\x32\x13.OSMPBF.StringTable\x12.\n\x0eprimitivegroup\x18\x02 \x03(\x0b\x32\x16.OSMPBF.PrimitiveGroup\x12\x18\n\x0bgranularity\x18\x11 \x01(\x05:\x03\x31\x30\x30\x12\x15\n\nlat_offset\x18\x13 \x01(\x03:\x01\x30\x12\x15\n\nlon_offset\x18\x14 \x01(\x03:\x01\x30\x12\x1e\n\x10\x64\x61te_granularity\x18\x12 \x01(\x05:\x04\x31\x30\x30\x30"\xb7\x01\n\x0ePrimitiveGroup\x12\x1b\n\x05nodes\x18\x01 \x03(\x0b\x32\x0c.OSMPBF.Node\x12!\n\x05\x64\x65nse\x18\x02 \x01(\x0b\x32\x12.OSMPBF.DenseNodes\x12\x19\n\x04ways\x18\x03 \x03(\x0b\x32\x0b.OSMPBF.Way\x12#\n\trelations\x18\x04 \x03(\x0b\x32\x10.OSMPBF.Relation\x12%\n\nchangesets\x18\x05 \x03(\x0b\x32\x11.OSMPBF.ChangeSet"\x18\n\x0bStringTable\x12\t\n\x01s\x18\x01 \x03(\x0c"q\n\x04Info\x12\x13\n\x07version\x18\x01 \x01(\x05:\x02-1\x12\x11\n\ttimestamp\x18\x02 \x01(\x03\x12\x11\n\tchangeset\x18\x03 \x01(\x03\x12\x0b\n\x03uid\x18\x04 \x01(\x05\x12\x10\n\x08user_sid\x18\x05 \x01(\r\x12\x0f\n\x07visible\x18\x06 \x01(\x08"\x8a\x01\n\tDenseInfo\x12\x13\n\x07version\x18\x01 \x03(\x05\x42\x02\x10\x01\x12\x15\n\ttimestamp\x18\x02 \x03(\x12\x42\x02\x10\x01\x12\x15\n\tchangeset\x18\x03 \x03(\x12\x42\x02\x10\x01\x12\x0f\n\x03uid\x18\x04 \x03(\x11\x42\x02\x10\x01\x12\x14\n\x08user_sid\x18\x05 \x03(\x11\x42\x02\x10\x01\x12\x13\n\x07visible\x18\x06 \x03(\x08\x42\x02\x10\x01"\x17\n\tChangeSet\x12\n\n\x02id\x18\x01 \x02(\x03"l\n\x04Node\x12\n\n\x02id\x18\x01 \x02(\x12\x12\x10\n\x04keys\x18\x02 \x03(\rB\x02\x10\x01\x12\x10\n\x04vals\x18\x03 \x03(\rB\x02\x10\x01\x12\x1a\n\x04info\x18\x04 \x01(\x0b\x32\x0c.OSMPBF.Info\x12\x0b\n\x03lat\x18\x08 \x02(\x12\x12\x0b\n\x03lon\x18\t \x02(\x12"{\n\nDenseNodes\x12\x0e\n\x02id\x18\x01 \x03(\x12\x42\x02\x10\x01\x12$\n\tdenseinfo\x18\x05 \x01(\x0b\x32\x11.OSMPBF.DenseInfo\x12\x0f\n\x03lat\x18\x08 \x03(\x12\x42\x02\x10\x01\x12\x0f\n\x03lon\x18\t \x03(\x12\x42\x02\x10\x01\x12\x15\n\tkeys_vals\x18\n \x03(\x05\x42\x02\x10\x01"c\n\x03Way\x12\n\n\x02id\x18\x01 \x02(\x03\x12\x10\n\x04keys\x18\x02 \x03(\rB\x02\x10\x01\x12\x10\n\x04vals\x18\x03 \x03(\rB\x02\x10\x01\x12\x1a\n\x04info\x18\x04 \x01(\x0b\x32\x0c.OSMPBF.Info\x12\x10\n\x04refs\x18\x08 \x03(\x12\x42\x02\x10\x01"\xe0\x01\n\x08Relation\x12\n\n\x02id\x18\x01 \x02(\x03\x12\x10\n\x04keys\x18\x02 \x03(\rB\x02\x10\x01\x12\x10\n\x04vals\x18\x03 \x03(\rB\x02\x10\x01\x12\x1a\n\x04info\x18\x04 \x01(\x0b\x32\x0c.OSMPBF.Info\x12\x15\n\troles_sid\x18\x08 \x03(\x05\x42\x02\x10\x01\x12\x12\n\x06memids\x18\t \x03(\x12\x42\x02\x10\x01\x12.\n\x05types\x18\n \x03(\x0e\x32\x1b.OSMPBF.Relation.MemberTypeB\x02\x10\x01"-\n\nMemberType\x12\x08\n\x04NODE\x10\x00\x12\x07\n\x03WAY\x10\x01\x12\x0c\n\x08RELATION\x10\x02\x42\x11\n\rcrosby.binaryH\x03'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "osmformat_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = b"\n\rcrosby.binaryH\003"
    _DENSEINFO.fields_by_name["version"]._options = None  # type: ignore[name-defined]
    _DENSEINFO.fields_by_name[
        "version"
    ]._serialized_options = b"\020\001"  # type: ifnore[name-defined]
    _DENSEINFO.fields_by_name["timestamp"]._options = None
    _DENSEINFO.fields_by_name["timestamp"]._serialized_options = b"\020\001"
    _DENSEINFO.fields_by_name["changeset"]._options = None
    _DENSEINFO.fields_by_name["changeset"]._serialized_options = b"\020\001"
    _DENSEINFO.fields_by_name["uid"]._options = None
    _DENSEINFO.fields_by_name["uid"]._serialized_options = b"\020\001"
    _DENSEINFO.fields_by_name["user_sid"]._options = None
    _DENSEINFO.fields_by_name["user_sid"]._serialized_options = b"\020\001"
    _DENSEINFO.fields_by_name["visible"]._options = None
    _DENSEINFO.fields_by_name["visible"]._serialized_options = b"\020\001"
    _NODE.fields_by_name["keys"]._options = None
    _NODE.fields_by_name["keys"]._serialized_options = b"\020\001"
    _NODE.fields_by_name["vals"]._options = None
    _NODE.fields_by_name["vals"]._serialized_options = b"\020\001"
    _DENSENODES.fields_by_name["id"]._options = None
    _DENSENODES.fields_by_name["id"]._serialized_options = b"\020\001"
    _DENSENODES.fields_by_name["lat"]._options = None
    _DENSENODES.fields_by_name["lat"]._serialized_options = b"\020\001"
    _DENSENODES.fields_by_name["lon"]._options = None
    _DENSENODES.fields_by_name["lon"]._serialized_options = b"\020\001"
    _DENSENODES.fields_by_name["keys_vals"]._options = None
    _DENSENODES.fields_by_name["keys_vals"]._serialized_options = b"\020\001"
    _WAY.fields_by_name["keys"]._options = None
    _WAY.fields_by_name["keys"]._serialized_options = b"\020\001"
    _WAY.fields_by_name["vals"]._options = None
    _WAY.fields_by_name["vals"]._serialized_options = b"\020\001"
    _WAY.fields_by_name["refs"]._options = None
    _WAY.fields_by_name["refs"]._serialized_options = b"\020\001"
    _RELATION.fields_by_name["keys"]._options = None
    _RELATION.fields_by_name["keys"]._serialized_options = b"\020\001"
    _RELATION.fields_by_name["vals"]._options = None
    _RELATION.fields_by_name["vals"]._serialized_options = b"\020\001"
    _RELATION.fields_by_name["roles_sid"]._options = None
    _RELATION.fields_by_name["roles_sid"]._serialized_options = b"\020\001"
    _RELATION.fields_by_name["memids"]._options = None
    _RELATION.fields_by_name["memids"]._serialized_options = b"\020\001"
    _RELATION.fields_by_name["types"]._options = None
    _RELATION.fields_by_name["types"]._serialized_options = b"\020\001"
    _globals["_HEADERBLOCK"]._serialized_start = 28
    _globals["_HEADERBLOCK"]._serialized_end = 291
    _globals["_HEADERBBOX"]._serialized_start = 293
    _globals["_HEADERBBOX"]._serialized_end = 363
    _globals["_PRIMITIVEBLOCK"]._serialized_start = 366
    _globals["_PRIMITIVEBLOCK"]._serialized_end = 576
    _globals["_PRIMITIVEGROUP"]._serialized_start = 579
    _globals["_PRIMITIVEGROUP"]._serialized_end = 762
    _globals["_STRINGTABLE"]._serialized_start = 764
    _globals["_STRINGTABLE"]._serialized_end = 788
    _globals["_INFO"]._serialized_start = 790
    _globals["_INFO"]._serialized_end = 903
    _globals["_DENSEINFO"]._serialized_start = 906
    _globals["_DENSEINFO"]._serialized_end = 1044
    _globals["_CHANGESET"]._serialized_start = 1046
    _globals["_CHANGESET"]._serialized_end = 1069
    _globals["_NODE"]._serialized_start = 1071
    _globals["_NODE"]._serialized_end = 1179
    _globals["_DENSENODES"]._serialized_start = 1181
    _globals["_DENSENODES"]._serialized_end = 1304
    _globals["_WAY"]._serialized_start = 1306
    _globals["_WAY"]._serialized_end = 1405
    _globals["_RELATION"]._serialized_start = 1408
    _globals["_RELATION"]._serialized_end = 1632
    _globals["_RELATION_MEMBERTYPE"]._serialized_start = 1587
    _globals["_RELATION_MEMBERTYPE"]._serialized_end = 1632
# @@protoc_insertion_point(module_scope)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions
from builtins import (
    bool,
    bytes,
    int,
    str,
)
from google.protobuf.descriptor import (
    Descriptor,
    FileDescriptor,
)
from google.protobuf.message import (
    Message,
)

DESCRIPTOR: FileDescriptor

@typing_extensions.final
class Blob(Message):
    """
    STORAGE LAYER: Storing primitives.
    """

    DESCRIPTOR: Descriptor

    RAW_FIELD_NUMBER: int
    RAW_SIZE_FIELD_NUMBER: int
    ZLIB_DATA_FIELD_NUMBER: int
    LZMA_DATA_FIELD_NUMBER: int
    OBSOLETE_BZIP2_DATA_FIELD_NUMBER: int
    raw: bytes
    """No compression"""
    raw_size: int
    """When compressed, the uncompressed size"""
    zlib_data: bytes
    """Possible compressed versions of the data."""
    lzma_data: bytes
    """PROPOSED feature for LZMA compressed data. SUPPORT IS NOT REQUIRED."""
    OBSOLETE_bzip2_data: bytes
    """Formerly used for bzip2 compressed data. Depreciated in 2010.
    Don't reuse this tag number.
    """
    def __init__(
        self,
        *,
        raw: bytes | None = ...,
        raw_size: int | None = ...,
        zlib_data: bytes | None = ...,
        lzma_data: bytes | None = ...,
        OBSOLETE_bzip2_data: bytes | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "OBSOLETE_bzip2_data",
            b"OBSOLETE_bzip2_data",
            "lzma_data",
            b"lzma_data",
            "raw",
            b"raw",
            "raw_size",
            b"raw_size",
            "zlib_data",
            b"zlib_data",
        ],
    ) -> bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "OBSOLETE_bzip2_data",
            b"OBSOLETE_bzip2_data",
            "lzma_data",
            b"lzma_data",
            "raw",
            b"raw",
            "raw_size",
            b"raw_size",
            "zlib_data",
            b"zlib_data",
        ],
    ) -> None: ...

@typing_extensions.final
class BlobHeader(Message):
    """A file contains an sequence of fileblock headers, each prefixed by
    their length in network byte order, followed by a data block
    containing the actual data. types staring with a "_" are reserved.
    """

    DESCRIPTOR: Descriptor

    TYPE_FIELD_NUMBER: int
    INDEXDATA_FIELD_NUMBER: int
    DATASIZE_FIELD_NUMBER: int
    type: str
    indexdata: bytes
    datasize: int
    def __init__(
        self,
        *,
        type: str | None = ...,
        indexdata: bytes | None = ...,
        datasize: int | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "datasize", b"datasize", "indexdata", b"indexdata", "type", b"type"
        ],
    ) -> bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "datasize", b"datasize", "indexdata", b"indexdata", "type", b"type"
        ],
    ) -> None: ...



from typing import Generator, Type

import geopandas as gpd
import numpy as np
import pandas as pd
from numpy.typing import NDArray
from shapely import Geometry, LinearRing, LineString, Polygon

from .unpacked import WayGroup


def infer_way_type(
    refs: NDArray[np.int64], tags: dict[str, str]
) -> Type[LinearRing] | Type[Polygon] | Type[LineString]:
    """Rules are taken from here: https://wiki.openstreetmap.org/wiki/Way#Types_of_way"""
    # if way is closed
    if refs[-1] == refs[0]:
        # exceptions where a closed way is not intended to be an area
        if "highway" not in tags and "barrier" not in tags:
            return LinearRing  # type: ignore[no-any-return]
        else:
            return Polygon  # type: ignore[no-any-return]
    else:
        return LineString  # type: ignore[no-any-return]


def _get_geometries(
    references: list[NDArray[np.int64]],
    tags: dict[int, dict[str, str]],
    nodes: gpd.GeoDataFrame,
) -> Generator[Geometry, None, None]:
    for i, refs in enumerate(references):
        geom_type = infer_way_type(refs, tags.get(i, {}))
        yield geom_type(nodes.loc[refs, "geometry"])


def consolidate_ways(group: WayGroup, nodes: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
    tags = pd.DataFrame.from_dict(
        group.tags,
        orient="index",
        dtype=str(pd.SparseDtype(str)),
    )

    return (
        gpd.GeoDataFrame(
            {
                "geometry": _get_geometries(group.member_ids, group.tags, nodes),
                "version": group.version,
                "changeset": group.changeset,
                "visible": group.visible,
                "id": group.ids,
            },
            crs="EPSG:4326",
        )
        .join(tags)
        .set_index("id")
    )



import struct
import zlib
from io import BufferedReader
from typing import Generator

from .proto import Blob, BlobHeader

__all__ = ["read_blocks"]


def read_blocks(f: BufferedReader) -> Generator[bytes, None, None]:
    """Return a generator of blocks of bytes from the given file.

    This corresponds to the fileformat part in the protobuf definitions.
    """
    while size_header := f.read(4):
        blob_header_size: int
        blob_header_size, *_ = struct.unpack("!L", size_header)

        blob: Blob = Blob.FromString(
            f.read(BlobHeader.FromString(f.read(blob_header_size)).datasize)
        )
        yield zlib.decompress(blob.zlib_data)



from __future__ import annotations

from dataclasses import dataclass
from itertools import accumulate

import numpy as np
from numpy.typing import NDArray

from osm4gpd.proto import PrimitiveGroup
from osm4gpd.tags import get_tags

from .base import BaseGroup

__all__ = ["WayGroup"]


@dataclass(repr=False)
class WayGroup(BaseGroup):
    member_ids: list[NDArray[np.int64]]

    @classmethod
    def from_primitive_group(
        cls, group: PrimitiveGroup, string_table: list[str]
    ) -> WayGroup:
        ids: list[int] = []
        versions: list[int] = []
        member_ids: list[NDArray[np.int64]] = []
        tags: dict[int, dict[str, str]] = {}
        visible: list[bool] = []
        changeset: list[int] = []

        for i, way in enumerate(group.ways):
            member_ids.append(np.fromiter(accumulate(way.refs), dtype=np.int64))
            _tags = get_tags(way, string_table)
            if len(_tags) > 0:
                tags[i] = _tags

            ids.append(way.id)
            # fixme: add optional here
            versions.append(way.info.version)
            visible.append(way.info.visible)
            changeset.append(way.info.changeset)

        return cls(
            ids=np.array(ids),
            tags=tags,
            member_ids=member_ids,
            version=versions,
            changeset=changeset,
            visible=visible,
        )



import geopandas as gpd
import pandas as pd

from .unpacked import NodesGroup


def consolidate_nodes(group: NodesGroup) -> gpd.GeoDataFrame:
    nodes = gpd.GeoDataFrame(
        {
            "geometry": gpd.points_from_xy(group.lon, group.lat, crs="EPSG:4326"),
            "id": group.ids,
            "version": group.version,
            "visible": group.visible,
            "changeset": group.changeset,
        }
    )

    tags = pd.DataFrame.from_dict(
        group.tags,
        orient="index",
        dtype=str(pd.SparseDtype(str)),
    )

    return nodes.join(tags).set_index("id")
